﻿using Newtonsoft.Json;
using GlaxOSC;

namespace VRC_Console;

public partial class OSCCollar
{

    public static void Main(string[] args)
    {
        _ = new OSCCollar();
    }

    private static readonly Config Config = File.Exists(Config.ConfigFilePath) ? JsonConvert.DeserializeObject<Config>(File.ReadAllText(Config.ConfigFilePath))! : new Config();
    private readonly OSC _osc;
    
    private Task _consoleOutputTask;
    
    private string _avatarId = "";
    private float _leashStretch;
    private readonly Vector _unitVectorLeash = new(1, 0);
    private Vector _movementVector = new(0, 0);
    private bool _allowMoving = true;
    private DateTime _lastNilMessageSent = DateTime.UnixEpoch;
    private DateTime _lastMessageReceived = DateTime.UnixEpoch;
    private string _lastMessageReceivedParameter = "null";
    private readonly DateTime _programStarted = DateTime.Now;
    private static readonly TimeSpan UpdateInterval = TimeSpan.FromMilliseconds(10); //Minimum is 10ms for VRC to register
    private static readonly TimeSpan MessageMinInterval = TimeSpan.FromMilliseconds(400); //Reset OSC movement at least once per second, to prevent hands from resetting to desktop position
    private ReferencePosition refPos1 = null!, refPos2 = null!, refPos3 = null!;
    private readonly Queue<ValueTuple<double, double>> _calibrationAverage = new();
    private double _constantA, _constantB, _constantC, _constantD, _constantE;  

    private OSCCollar(bool skipSetup = true)
    {
        this._osc = new OSC("OSCCollar", new List<OSCEndpoint>()
        {
            new("/avatar/parameters/GPS1", typeof(float), OnGPSParameterChanged),
            new("/avatar/parameters/GPS2", typeof(float), OnGPSParameterChanged),
            new("/avatar/parameters/GPS3", typeof(float), OnGPSParameterChanged),
            new($"/avatar/parameters/{Config.LeashStretchParameter}", typeof(float), OnLeashStretchChanged),
            new($"/avatar/parameters/{Config.LeashToggleParameter}", typeof(bool), OnLeashToggled),
        });
        this._osc.OnParameterChangeEvent += OnOscParameterChangeEvent;
        
        this.SetupGPSVars();
        if (!skipSetup)
        {
            Console.WriteLine(Resources.OSCCollar_OSCCollar_Position_your_GPS_receivers);
            Console.WriteLine(Resources.OSCCollar_OSCCollar_GPS_1_position, refPos1.X, refPos1.Y);
            Console.WriteLine(Resources.OSCCollar_OSCCollar_GPS_2_position, refPos2.X, refPos2.Y);
            Console.WriteLine(Resources.OSCCollar_OSCCollar_GPS_3_position, refPos3.X, refPos3.Y);
            Console.WriteLine(Resources.OSCCollar_OSCCollar_Radius_of_each_receiver, Config.Radius * 2);
            Console.ReadKey();
        }
        
        Thread runningThread = new(RunningThread);
        runningThread.Start();
    }

    private void OnLeashToggled(string endpoint, object? oldValue, object? newValue)
    {
        this._allowMoving = (bool)(newValue ?? false);
    }


    private void OnLeashStretchChanged(string endpoint, object? oldValue, object? newValue)
    {
        this._leashStretch = (float)(newValue ?? 0);
    }

    private void OnGPSParameterChanged(string endpoint, object? oldValue, object? newValue)
    {
        switch (endpoint)
        {
            case "/avatar/parameters/GPS1":
                this.refPos1.Distance = 1 - (float)(newValue ?? 1);
                break;
            case "/avatar/parameters/GPS2":
                this.refPos2.Distance = 1 - (float)(newValue ?? 1);
                break;
            case "/avatar/parameters/GPS3":
                this.refPos3.Distance = 1 - (float)(newValue ?? 1);
                break;
        }
    }


    private void OnOscParameterChangeEvent(string endpoint, object? oldValue, object? newValue)
    {
        _lastMessageReceived = DateTime.Now;
        _lastMessageReceivedParameter = endpoint;
        switch (endpoint)
        {
            case "/avatar/change":
                this._avatarId = (string)(newValue ?? "");
                break;
        }
    }

    #region Setup
    private void SetupGPSVars()
    {
        this.refPos1 = new(0, -Config.Radius);
        this.refPos2 = new(-(Config.Radius * 0.866), Config.Radius / 2);
        this.refPos3 = new(Config.Radius * 0.866, Config.Radius / 2);
        this._constantA = 2 * refPos2.X - 2 * refPos1.X;
        this._constantB = 2 * refPos2.Y - 2 * refPos1.Y;
        this._constantC = Math.Pow(refPos1.X, 2) + Math.Pow(refPos2.X, 2) - Math.Pow(refPos1.Y, 2) + Math.Pow(refPos2.Y, 2);
        this._constantD = 2 * refPos3.X - 2 * refPos2.X;
        this._constantE = Math.Pow(refPos2.X, 2) + Math.Pow(refPos3.X, 2) - Math.Pow(refPos2.Y, 2) + Math.Pow(refPos3.Y, 2);
    }
    #endregion

    private void CalculatePositionFromGPS()
    {
        if (this.refPos1.Distance == 0 || this.refPos2.Distance == 0 || this.refPos3.Distance == 0)
            return;
        double factorA = Math.Pow(refPos1.Distance, 2) - Math.Pow(refPos2.Distance, 2) - _constantC;
        double factorB = Math.Pow(refPos2.Distance, 2) - Math.Pow(refPos3.Distance, 2) - _constantE;

        double calculatedPosX = (factorB / _constantD);
        double calculatedPosY = -1 * (factorA * _constantD - _constantA * factorB) / (_constantB * _constantD);

        if (this._leashStretch == 0)
            AddCalibrationValues(calculatedPosX, calculatedPosY);
        
        ValueTuple<double, double> calibrationValues = GetCalibrationValues();
        
        double calibratedPosX = calculatedPosX + calibrationValues.Item1;
        double calibratedPosY = calculatedPosY + calibrationValues.Item2;
        
        double inverseLength = 1 / Math.Sqrt(Math.Pow(calibratedPosX, 2) + Math.Pow(calibratedPosY, 2));

        _unitVectorLeash.X = inverseLength is not Double.PositiveInfinity ? calibratedPosX * inverseLength : 0;
        _unitVectorLeash.Y = inverseLength is not Double.PositiveInfinity ? calibratedPosY * inverseLength : 0;
        
        if (_leashStretch < Config.WalkStretchDeadzone) //Below Deadzone
            _movementVector = new();
        else if (_leashStretch < Config.RunStretch) //Walk
            _movementVector = _unitVectorLeash.MultipliedWith(0.5);
        else //Run
            _movementVector = _unitVectorLeash.MultipliedWith(1);
    }

    private void RunningThread()
    {
        while (true)
        {
            CalculatePositionFromGPS();
            if (_lastNilMessageSent.Add(MessageMinInterval) < DateTime.Now)
            {
                this._osc.Client.Send("/input/Vertical", 0f);
                this._osc.Client.Send("/input/Horizontal", 0f);
                this._lastNilMessageSent = DateTime.Now;
            }else if (_allowMoving)
            {
                this._osc.Client.Send("/input/Vertical", Convert.ToSingle(_movementVector.Y));
                this._osc.Client.Send("/input/Horizontal", Convert.ToSingle(_movementVector.X));
            }
            
            PrintConsole();
            Thread.Sleep(UpdateInterval);
        }
    }

    private void AddCalibrationValues(double x, double y)
    {
        this._calibrationAverage.Enqueue(new ValueTuple<double, double>(x, y));
        while (this._calibrationAverage.Count > Config.CalibrationValues)
            this._calibrationAverage.Dequeue();
    }

    private ValueTuple<double, double> GetCalibrationValues()
    {
        if (this._calibrationAverage.Count < 1)
            return new ValueTuple<double, double>(0, 0);
        double averageX = -this._calibrationAverage.Average(value => value.Item1);
        double averageY = -this._calibrationAverage.Average(value => value.Item2);
        return new ValueTuple<double, double>(averageX, averageY);
    }
}